@@ -31,11 +31,12 @@ class CloneDialog(QDialog):
         self.ret = None
 
         dest = src = cwd = os.getcwd()
-        if len(args) > 1:
-            src = args[0]
-            dest = args[1]
-        elif len(args):
-            src = args[0]
+        if args:
+            if len(args) > 1:
+                src = args[0]
+                dest = args[1]
+            else:
+                src = args[0]
         udest = hglib.tounicode(dest)
         usrc = hglib.tounicode(src)
         ucwd = hglib.tounicode(cwd)
@@ -73,22 +74,20 @@ class CloneDialog(QDialog):
         grid.addWidget(self.dest_combo, 1, 1)
         grid.addWidget(self.dest_btn, 1, 2)
 
-        # workaround for QComboBox to test edited text case-sensitively
-        # see src/gui/widgets/qcombobox.cpp:
-        # QComboBoxPrivate::_q_editingFinished() and matchFlags()
-        for name in ('src_combo', 'dest_combo'):
-            w = getattr(self, name).lineEdit()
-            w.completer().setCaseSensitivity(Qt.CaseSensitive)
+        for combo in (self.src_combo, self.dest_combo):
+            qtlib.allowCaseChangingInput(combo)
 
         s = QSettings()
         self.shist = s.value('clone/source').toStringList()
         for path in self.shist:
             if path: self.src_combo.addItem(path)
+        self.src_combo.setCurrentIndex(-1)
         self.src_combo.setEditText(usrc)
 
         self.dhist = s.value('clone/dest').toStringList()
         for path in self.dhist:
             if path: self.dest_combo.addItem(path)
+        self.dest_combo.setCurrentIndex(-1)
         self.dest_combo.setEditText(udest)
 
         ### options
@@ -221,7 +220,7 @@ class CloneDialog(QDialog):
         rev = opts.get('rev')
         if rev:
             self.rev_chk.setChecked(True)
-            self.rev_text.setText(hglib.tounicode(', '.join(rev)))
+            self.rev_text.setText(hglib.tounicode(rev))
         self.noupdate_chk.setChecked(bool(opts.get('noupdate')))
         self.pproto_chk.setChecked(bool(opts.get('pull')))
         self.uncomp_chk.setChecked(bool(opts.get('uncompressed')))
