@@ -456,7 +456,7 @@ class _Entry(object):
         for e in self._child.itervalues():
             e.sort(reverse=reverse)
         self._nameindex.sort(
-            key=lambda s: '%s%s' % (self[s] and 'D' or 'F', s),
+            key=lambda s: '%s%s' % (self[s] and 'D' or 'F', os.path.normcase(s)),
             reverse=reverse)
 
 class ManifestCompleter(QCompleter):
