@@ -7,7 +7,7 @@
 # GNU General Public License version 2, incorporated herein by reference.
 
 shortlicense = '''
-Copyright (C) 2008-2012 Steve Borho <steve@borho.org> and others.
+Copyright (C) 2008-2013 Steve Borho <steve@borho.org> and others.
 This is free software; see the source for copying conditions.  There is NO
 warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 '''
@@ -29,8 +29,7 @@ from mercurial import util, fancyopts, cmdutil, extensions, error, scmutil
 from tortoisehg.hgqt.i18n import agettext as _
 from tortoisehg.util import hglib, paths, i18n
 from tortoisehg.util import version as thgversion
-from tortoisehg.hgqt import qtlib
-from tortoisehg.hgqt.bugreport import run as bugrun
+from tortoisehg.hgqt import bugreport, qtlib
 
 try:
     from tortoisehg.util.config import nofork as config_nofork
@@ -63,9 +62,10 @@ def dispatch(args):
         opts['nofork'] = True
         errstring = _('Error string "%(arg0)s" at %(arg1)s<br>Please '
                       '<a href="#edit:%(arg1)s">edit</a> your config')
-        main = QApplication(sys.argv)
-        dlg = ExceptionMsgBox(hglib.tounicode(str(e)), errstring, opts,
-                              parent=None)
+        if not QApplication.instance():
+            main = QApplication(sys.argv)
+        dlg = ExceptionMsgBox(hglib.tounicode(str(e)),
+                              hglib.tounicode(errstring), opts, parent=None)
         dlg.exec_()
     except SystemExit:
         pass
@@ -76,8 +76,11 @@ def dispatch(args):
         opts = {}
         opts['cmd'] = ' '.join(sys.argv[1:])
         opts['error'] = traceback.format_exc()
-        opts['nofork'] = True
-        return qtrun(bugrun, u, **opts)
+        if not QApplication.instance():
+            main = QApplication(sys.argv)
+        dlg = bugreport.BugReport(opts)
+        dlg.exec_()
+        return -1
     except KeyboardInterrupt:
         print _('\nCaught keyboard interrupt, aborting.\n')
 
@@ -220,7 +223,7 @@ def _parse(ui, args):
         c.append((o[0], o[1], options[o[1]], o[3]))
 
     try:
-        args = fancyopts.fancyopts(args, c, cmdoptions)
+        args = fancyopts.fancyopts(args, c, cmdoptions, True)
     except fancyopts.getopt.GetoptError, inst:
         raise error.CommandError(cmd, inst)
 
@@ -502,8 +505,8 @@ class _QtRunner(QObject):
                                   hglib.tounicode(errstr), opts,
                                   parent=self._mainapp.activeWindow())
         elif etype is KeyboardInterrupt:
-            if qtlib.QuestionMsgBox(_('Keyboard interrupt'),
-                    _('Close this application?')):
+            if qtlib.QuestionMsgBox(hglib.tounicode(_('Keyboard interrupt')),
+                    hglib.tounicode(_('Close this application?'))):
                 QApplication.quit()
             else:
                 self.errors = []
@@ -526,13 +529,14 @@ class _QtRunner(QObject):
         QSettings.setDefaultFormat(QSettings.IniFormat)
 
         self._mainapp = QApplication(sys.argv)
-        self._gc = GarbageCollector(self, self.debug)
         try:
+            self._gc = GarbageCollector(self, self.debug)
             # default org is used by QSettings
             self._mainapp.setApplicationName('TortoiseHgQt')
             self._mainapp.setOrganizationName('TortoiseHg')
             self._mainapp.setOrganizationDomain('tortoisehg.org')
             self._mainapp.setApplicationVersion(thgversion.version())
+            self._fixlibrarypaths()
             self._installtranslator()
             qtlib.setup_font_substitutions()
             qtlib.fix_application_font()
@@ -548,13 +552,15 @@ class _QtRunner(QObject):
                     from tortoisehg.hgqt import thgrepo
                     thgrepo.repository(ui, opts['repository'])
                 except error.RepoError, e:
-                    qtlib.WarningMsgBox(_('Repository Error'),
+                    qtlib.WarningMsgBox(hglib.tounicode(_('Repository Error')),
                                         hglib.tounicode(str(e)))
                     return
             dlg = dlgfunc(ui, *args, **opts)
             if dlg:
                 dlg.show()
                 dlg.raise_()
+            else:
+                return -1
         except:
             # Exception before starting eventloop needs to be postponed;
             # otherwise it will be ignored silently.
@@ -570,6 +576,12 @@ class _QtRunner(QObject):
         finally:
             self._mainapp = None
 
+    def _fixlibrarypaths(self):
+        # make sure to use the bundled Qt plugins to avoid ABI incompatibility
+        # http://qt-project.org/doc/qt-4.8/deployment-windows.html#qt-plugins
+        if os.name == 'nt' and getattr(sys, 'frozen', False):
+            self._mainapp.setLibraryPaths([self._mainapp.applicationDirPath()])
+
     def _installtranslator(self):
         if not i18n.language:
             return
@@ -637,6 +649,11 @@ def email(ui, *pats, **opts):
     from tortoisehg.hgqt.hgemail import run
     return qtrun(run, ui, *pats, **opts)
 
+def graft(ui, *revs, **opts):
+    """graft dialog"""
+    from tortoisehg.hgqt.graft import run
+    return qtrun(run, ui, *revs, **opts)
+
 def resolve(ui, *pats, **opts):
     """resolve dialog"""
     from tortoisehg.hgqt.resolve import run
@@ -692,10 +709,9 @@ def mq(ui, *pats, **opts):
     from tortoisehg.hgqt.mq import run
     return qtrun(run, ui, *pats, **opts)
 
-def test(ui, *pats, **opts):
-    """test arbitrary widgets"""
-    from tortoisehg.hgqt.mq import run
-    return qtrun(run, ui, *pats, **opts)
+def debugbugreport(ui, *pats, **opts):
+    """open bugreport dialog by exception"""
+    raise Exception(' '.join(pats))
 
 def purge(ui, *pats, **opts):
     """purge unknown and/or ignore files from repository"""
@@ -828,6 +844,11 @@ def thgimport(ui, *pats, **opts):
     from tortoisehg.hgqt.thgimport import run
     return qtrun(run, ui, *pats, **opts)
 
+def revdetails(ui, *pats, **opts):
+    """revision details tool"""
+    from tortoisehg.hgqt.revdetails import run
+    return qtrun(run, ui, *pats, **opts)
+
 ### help management, adapted from mercurial.commands.help_()
 def help_(ui, name=None, with_version=False, **opts):
     """show help for a command, extension, or list of commands
@@ -1074,7 +1095,7 @@ table = {
              '(only a repository)')),
           ('u', 'updaterev', '',
            _('revision, tag or branch to check out')),
-          ('r', 'rev', [], _('include the specified changeset')),
+          ('r', 'rev', '', _('include the specified changeset')),
           ('b', 'branch', [],
            _('clone only the specified branch')),
           ('', 'pull', None, _('use pull protocol to copy metadata')),
@@ -1087,6 +1108,9 @@ table = {
         _('thg commit [OPTIONS] [FILE]...')),
     "drag_move": (drag_move, [], _('thg drag_move SOURCE... DEST')),
     "drag_copy": (drag_copy, [], _('thg drag_copy SOURCE... DEST')),
+    "graft": (graft,
+        [('r', 'rev', [], _('revisions to graft'))],
+        _('thg graft [-r] REV...')),
     "^grep|search": (grep,
         [('i', 'ignorecase', False, _('ignore case during search')),],
         _('thg grep')),
@@ -1102,8 +1126,12 @@ table = {
          _('thg email [REVS]')),
     "^log|history|explorer|workbench":
         (log,
-         [('l', 'limit', '', _('limit number of changes displayed'))],
+         [('l', 'limit', '', _('(DEPRECATED)'))],
          _('thg log [OPTIONS] [FILE]')),
+    "^revdetails":
+        (revdetails,
+         [('r', 'rev', '', _('the revision to show'))],
+         _('thg revdetails [-r REV]')),
     "manifest":
         (manifest,
          [('r', 'rev', '', _('revision to display')),
@@ -1127,7 +1155,7 @@ table = {
           ('', 'webdir-conf', '',
            _('name of the hgweb config file (DEPRECATED)'))],
          _('thg serve [--web-conf FILE]')),
-    "^sync|synchronize": (sync, [], _('thg sync')),
+    "^sync|synchronize": (sync, [], _('thg sync [PEER]')),
     "^status|st": (status,
          [('c', 'clean', False, _('show files without changes')),
           ('i', 'ignored', False, _('show ignored files'))],
@@ -1140,13 +1168,12 @@ table = {
     "^rebase": (rebase,
         [('', 'keep', False, _('keep original changesets')),
          ('', 'keepbranches', False, _('keep original branch names')),
-         ('', 'detach', False, _('force detaching of source from its original '
-                                'branch')),
+         ('', 'detach', False, _('(DEPRECATED)')),
          ('s', 'source', '',
           _('rebase from the specified changeset')),
          ('d', 'dest', '',
           _('rebase onto the specified changeset'))],
-        _('thg rebase -s REV -d REV [--keep] [--detach]')),
+        _('thg rebase -s REV -d REV [--keep]')),
     "^tag":
         (tag,
          [('f', 'force', None, _('replace existing tag')),
@@ -1155,9 +1182,17 @@ table = {
           ('', 'remove', None, _('remove a tag')),
           ('m', 'message', '', _('use <text> as commit message')),],
          _('thg tag [-f] [-l] [-m TEXT] [-r REV] [NAME]')),
+    "thgstatus": (thgstatus,
+        [('',  'delay', None, _('wait until the second ticks over')),
+         ('n', 'notify', [], _('notify the shell for paths given')),
+         ('',  'remove', None, _('remove the status cache')),
+         ('s', 'show', None, _('show the contents of the '
+                               'status cache (no update)')),
+         ('',  'all', None, _('udpate all repos in current dir')) ],
+        _('thg thgstatus [OPTION]')),
     "shelve|unshelve": (shelve, [], _('thg shelve')),
     "rejects": (rejects, [], _('thg rejects [FILE]')),
-    "test": (test, [], _('thg test')),
+    "debugbugreport": (debugbugreport, [], _('thg debugbugreport [TEXT]')),
     "help": (help_, [], _('thg help [COMMAND]')),
     "^purge": (purge, [], _('thg purge')),
     "^qreorder": (qreorder, [], _('thg qreorder')),
